from pydantic import BaseModel
from tierkreis.controller.data.location import Loc
from tierkreis.controller.storage.protocol import ControllerStorage
from tierkreis import Labels

from tierkreis_visualization.data.eval import check_error
from tierkreis_visualization.data.models import PyNode, PyEdge


class LoopNodeData(BaseModel):
    nodes: list[PyNode]
    edges: list[PyEdge]


def get_loop_node(
    storage: ControllerStorage, node_location: Loc, errored_nodes: list[Loc]
) -> LoopNodeData:
    i = 0
    while storage.is_node_started(node_location.L(i + 1)):
        i += 1
    new_location = node_location.L(i)

    nodes = [PyNode(id=n, status="Finished", function_name=f"L{n}") for n in range(i)]

    if check_error(node_location, errored_nodes):
        last_status = "Error"
    elif storage.is_node_finished(new_location):
        last_status = "Finished"
    else:
        last_status = "Started"
    nodes.append(PyNode(id=i, status=last_status, function_name=f"L{i}"))

    edges = [
        PyEdge(from_node=n, from_port=Labels.VALUE, to_node=n + 1, to_port=Labels.VALUE)
        for n in range(i)
    ]

    return LoopNodeData(nodes=nodes, edges=edges)



from typing import Literal
from pydantic import BaseModel
from tierkreis.controller.data.graph import PortID

NodeStatus = Literal["Not started", "Started", "Error", "Finished"]


class PyNode(BaseModel):
    id: int | PortID
    status: NodeStatus
    function_name: str


class PyEdge(BaseModel):
    from_node: int
    from_port: str
    to_node: int
    to_port: str



import os
from uuid import UUID

from pydantic import BaseModel
from tierkreis.controller.data.location import Loc
from tierkreis_visualization.config import CONFIG, get_storage


class WorkflowDisplay(BaseModel):
    id: UUID
    id_int: int
    name: str | None


def get_workflows() -> list[WorkflowDisplay]:
    folders = os.listdir(CONFIG.tierkreis_path)
    folders.sort()
    workflows: list[WorkflowDisplay] = []
    for folder in folders:
        try:
            id = UUID(folder)
            metadata = get_storage(id).read_metadata(Loc(""))
            name = metadata["name"] or "workflow"
            workflows.append(WorkflowDisplay(id=id, id_int=int(id), name=name))
        except (TypeError, ValueError):
            continue

    return workflows



from uuid import UUID

from tierkreis.controller.data.location import Loc


def breadcrumb_links(crumbs: list[str]) -> list[tuple[str, str]]:
    url_path = ""
    links: list[tuple[str, str]] = []
    for crumb in crumbs:
        url_path = url_path + crumb
        links.append((crumb, url_path))

    return links


def breadcrumbs(workflow_id: UUID, node_location: Loc) -> list[tuple[str, str]]:
    node_location_strs: list[str] = [f".{x}" for x in node_location.split(".")]
    static_links = ["/workflows", f"/{workflow_id}/nodes/-"]
    return breadcrumb_links(static_links + node_location_strs[1:])



from pydantic import BaseModel
from tierkreis.controller.data.location import Loc
from tierkreis.controller.storage.protocol import ControllerStorage
from tierkreis.exceptions import TierkreisError


class FunctionDefinition(BaseModel):
    has_error: bool = False
    error_message: str | None = None


def get_function_node(storage: ControllerStorage, loc: Loc) -> FunctionDefinition:
    parent = loc.parent()
    if parent is None:
        raise TierkreisError("Func node must have parent.")
    if not storage.node_has_error(loc):
        return FunctionDefinition()
    return FunctionDefinition(has_error=True, error_message=storage.read_errors(loc))



from pathlib import Path


from fastapi import FastAPI, Request
from fastapi.staticfiles import StaticFiles
from fastapi.responses import RedirectResponse
import uvicorn

from tierkreis_visualization.routers.workflows import router as workflows_router

app = FastAPI()

app.include_router(workflows_router)
app.mount(
    "/static",
    StaticFiles(directory=(Path(__file__).parent.parent / "static").absolute()),
    name="static",
)


@app.get("/")
def read_root(request: Request):
    return RedirectResponse(url="/workflows")


def start():
    uvicorn.run(app)



import json
from typing import Optional, assert_never

from pydantic import BaseModel
from tierkreis.controller.data.location import WorkerCallArgs, Loc
from tierkreis.controller.data.graph import GraphData
from tierkreis.controller.storage.adjacency import in_edges
from tierkreis.controller.storage.protocol import ControllerStorage

from tierkreis_visualization.data.models import PyNode, NodeStatus, PyEdge


class EvalNodeData(BaseModel):
    nodes: list[PyNode]
    edges: list[PyEdge]


def node_status(
    is_finished: bool, definition: Optional[WorkerCallArgs], has_error: bool = False
) -> NodeStatus:
    if is_finished:
        return "Finished"

    if definition is not None:
        if has_error:
            return "Error"
        return "Started"

    return "Not started"


def check_error(node_location: Loc, errored_nodes: list[Loc]) -> bool:
    return any(node.startswith(node_location) for node in errored_nodes)


def get_eval_node(
    storage: ControllerStorage, node_location: Loc, errored_nodes: list[Loc]
) -> EvalNodeData:
    thunk = storage.read_output(node_location.N(-1), "body")
    graph = GraphData(**json.loads(thunk))

    pynodes: list[PyNode] = []
    for i, node in enumerate(graph.nodes):
        new_location = node_location.N(i)
        is_finished = storage.is_node_finished(new_location)
        has_error = check_error(new_location, errored_nodes)
        try:
            definition = storage.read_worker_call_args(new_location)
        except FileNotFoundError:
            definition = None

        status = node_status(is_finished, definition, has_error)

        match node.type:
            case "function":
                name = node.function_name
            case "const" | "map" | "eval" | "input" | "output" | "loop" | "ifelse":
                name = node.type
            case _:
                assert_never(node)

        pynode = PyNode(id=i, status=status, function_name=name)
        pynodes.append(pynode)

    py_edges: list[PyEdge] = []
    for idx, node in enumerate(graph.nodes):
        for p0, (i, p1) in in_edges(node).items():
            py_edge = PyEdge(from_node=i, from_port=p1, to_node=idx, to_port=p0)
            py_edges.append(py_edge)

    return EvalNodeData(nodes=pynodes, edges=py_edges)



import json
from typing import Any
from uuid import UUID

from fastapi import APIRouter, Request
from fastapi.responses import StreamingResponse
from pydantic import BaseModel
from starlette.responses import JSONResponse, PlainTextResponse
from tierkreis.controller.data.location import Loc, WorkerCallArgs
from watchfiles import awatch

from tierkreis_visualization.config import CONFIG, get_storage, templates
from tierkreis_visualization.data.eval import get_eval_node
from tierkreis_visualization.data.function import get_function_node
from tierkreis_visualization.data.loop import get_loop_node
from tierkreis_visualization.data.map import get_map_node
from tierkreis_visualization.data.workflows import get_workflows
from tierkreis_visualization.routers.models import PyGraph
from tierkreis_visualization.routers.navigation import breadcrumbs

router = APIRouter(prefix="/workflows")


@router.get("/")
def list_workflows(request: Request):
    workflows = get_workflows()
    return templates.TemplateResponse(
        request=request, name="workflows.html", context={"workflows": workflows}
    )


class NodeResponse(BaseModel):
    definition: WorkerCallArgs


def parse_node_location(node_location_str: str) -> Loc:
    return Loc(node_location_str)


def get_errored_nodes(workflow_id: UUID) -> list[Loc]:
    storage = get_storage(workflow_id)
    errored_nodes = storage.read_errors(Loc("-"))
    return [parse_node_location(node) for node in errored_nodes.split("\n")]


def get_node_data(workflow_id: UUID, node_location: Loc) -> dict[str, Any]:
    storage = get_storage(workflow_id)
    errored_nodes = get_errored_nodes(workflow_id)

    try:
        definition = storage.read_worker_call_args(node_location)
    except FileNotFoundError:
        return {
            "breadcrumbs": breadcrumbs(workflow_id, node_location),
            "url": f"/workflows/{workflow_id}/nodes/{node_location}",
            "node_location": str(node_location),
            "name": "unavailable.jinja",
        }

    node = storage.read_node_def(node_location)

    if node.type == "eval":
        data = get_eval_node(storage, node_location, errored_nodes)
        name = "eval.jinja"
        ctx: dict[str, Any] = PyGraph(nodes=data.nodes, edges=data.edges).model_dump(
            by_alias=True
        )

    elif node.type == "loop":
        data = get_loop_node(storage, node_location, errored_nodes)
        name = "loop.jinja"
        ctx = PyGraph(nodes=data.nodes, edges=data.edges).model_dump(by_alias=True)

    elif node.type == "map":
        data = get_map_node(storage, node_location, node, errored_nodes)
        name = "map.jinja"
        ctx = PyGraph(nodes=data.nodes, edges=data.edges).model_dump(by_alias=True)

    elif node.type == "function":
        data = get_function_node(storage, node_location)
        name = "function.jinja"
        ctx = {"definition": definition.model_dump(), "data": data}

    else:
        name = "fallback.html"
        ctx = {"definition": definition.model_dump()}

    ctx["breadcrumbs"] = breadcrumbs(workflow_id, node_location)
    ctx["url"] = f"/workflows/{workflow_id}/nodes/{node_location}"
    ctx["node_location"] = str(node_location)
    ctx["name"] = name

    return ctx


async def node_stream(workflow_id: UUID, node_location: Loc):
    node_path = CONFIG.tierkreis_path / str(workflow_id) / str(node_location)
    async for _changes in awatch(node_path, recursive=False):
        if (node_path / "definition").exists():
            ctx = get_node_data(workflow_id, node_location)
            yield f"event: message\ndata: {json.dumps(ctx)}\n\n"


@router.get("/{workflow_id}/nodes/{node_location_str}")
def get_node(request: Request, workflow_id: UUID, node_location_str: str):
    node_location = parse_node_location(node_location_str)
    ctx = get_node_data(workflow_id, node_location)
    if request.headers["Accept"] == "application/json":
        return JSONResponse(ctx)

    if request.headers["Accept"] == "text/event-stream":
        return StreamingResponse(
            node_stream(workflow_id, node_location), media_type="text/event-stream"
        )

    return templates.TemplateResponse(request=request, name=ctx["name"], context=ctx)


@router.get("/{workflow_id}/nodes/{node_location_str}/inputs/{port_name}")
def get_input(workflow_id: UUID, node_location_str: str, port_name: str):
    node_location = parse_node_location(node_location_str)
    storage = get_storage(workflow_id)
    definition = storage.read_worker_call_args(node_location)

    with open(definition.inputs[port_name], "rb") as fh:
        return JSONResponse(json.loads(fh.read()))


@router.get("/{workflow_id}/nodes/{node_location_str}/outputs/{port_name}")
def get_output(workflow_id: UUID, node_location_str: str, port_name: str):
    node_location = parse_node_location(node_location_str)
    storage = get_storage(workflow_id)
    definition = storage.read_worker_call_args(node_location)

    with open(definition.outputs[port_name], "rb") as fh:
        return JSONResponse(json.loads(fh.read()))


@router.get("/{workflow_id}/nodes/{node_location_str}/logs")
def get_logs(workflow_id: UUID, node_location_str: str):
    node_location = parse_node_location(node_location_str)
    storage = get_storage(workflow_id)
    definition = storage.read_worker_call_args(node_location)

    if definition.logs_path is None:
        return PlainTextResponse("Node definition not found.")

    messages = ""

    with open(definition.logs_path, "rb") as fh:
        for line in fh:
            messages += line.decode()

    return PlainTextResponse(messages)



from pathlib import Path
from uuid import UUID

from pydantic_settings import BaseSettings
from starlette.templating import Jinja2Templates
from tierkreis.controller.storage.protocol import ControllerStorage
from tierkreis.controller.storage.filestorage import ControllerFileStorage


class Settings(BaseSettings):
    tierkreis_path: Path = Path.home() / ".tierkreis" / "checkpoints"


CONFIG = Settings()

CONFIG.tierkreis_path.mkdir(parents=True, exist_ok=True)

templates = Jinja2Templates(
    directory=(Path(__file__).parent.parent / "templates").absolute()
)


def get_storage(workflow_id: UUID) -> ControllerStorage:
    return ControllerFileStorage(
        workflow_id=workflow_id, tierkreis_directory=CONFIG.tierkreis_path
    )



from pydantic import BaseModel
from tierkreis.controller.data.location import Loc
from tierkreis.controller.storage.protocol import ControllerStorage
from tierkreis.controller.data.graph import Map
from tierkreis.exceptions import TierkreisError

from tierkreis_visualization.data.eval import check_error
from tierkreis_visualization.data.models import PyEdge, PyNode


class MapNodeData(BaseModel):
    nodes: list[PyNode]
    edges: list[PyEdge]


def get_map_node(
    storage: ControllerStorage, loc: Loc, map: Map, errored_nodes: list[Loc]
) -> MapNodeData:
    parent = loc.parent()
    if parent is None:
        raise TierkreisError("MAP node must have parent.")

    map_eles = storage.read_output_ports(parent.N(map.input_idx))
    nodes: list[PyNode] = []
    for ele in map_eles:
        node = PyNode(id=ele, status="Started", function_name=ele)
        if check_error(loc.M(ele), errored_nodes):
            node.status = "Error"
        elif storage.is_node_finished(loc.M(ele)):
            node.status = "Finished"
        nodes.append(node)

    return MapNodeData(nodes=nodes, edges=[])
